#include "processing.h"
#include "fft_windows.h"

#include "arm_math.h"
#include "com/debug.h"

void samples_to_milliamps(float *floats, const uint32_t *uints, uint16_t count)
{
	// we'll convert it to millamps
//	dbg("Converting samples to milliamps...");

	for (int i = 0; i < count; i++) {
		const float digital = (float) uints[i];
		const float mv_adc = digital * (MV_RANGE / DIGITAL_RANGE);
		const float mv_shunt = mv_adc / PREAMP_FACTOR;
		const float ma_load = mv_shunt / CT_VOLTS_PER_AMP;

		floats[i] = ma_load;
	}

	// Now it's in milliamps, but with offset of half the scale
	// let's take mean value and align to that - will always be true for AC past a transformer
	// (note: there will be some offset caused by partially captured periods)

	// get mean
	float mean;
	arm_mean_f32(floats, count, &mean);

	for (int i = 0; i < count; i++) {
		floats[i] = (floats[i] - mean); // rougly approx. to real voltage
	}
}


void samples_to_fft(float *bins, const float *samples, uint16_t samp_count)
{
	assert_param(IS_FFT_LEN(samp_count));

	// Select the CFFT instance and a FFT window to use.

	const float *window;
	const arm_cfft_instance_f32 *S;

	switch (samp_count) {
		case 16:
			S = &arm_cfft_sR_f32_len16;
			window = win_hamming_16;
			break;
		case 32:
			S = &arm_cfft_sR_f32_len32;
			window = win_hamming_32;
			break;
		case 64:
			S = &arm_cfft_sR_f32_len64;
			window = win_hamming_64;
			break;
		case 128:
			S = &arm_cfft_sR_f32_len128;
			window = win_hamming_128;
			break;
		case 256:
			S = &arm_cfft_sR_f32_len256;
			window = win_hamming_256;
			break;
		case 512:
			S = &arm_cfft_sR_f32_len512;
			window = win_hamming_512;
			break;
		case 1024:
			S = &arm_cfft_sR_f32_len1024;
			window = win_hamming_1024;
			break;
		case 2048:
			S = &arm_cfft_sR_f32_len2048;
			window = win_hamming_2048;
			break;
//		case 4096:
//			S = &arm_cfft_sR_f32_len4096;
//			window = win_hamming_4096;
//			break;
		default:
			error("Bad FFT len %d.", samp_count);
			return;
	}

	// We have to convert the numbers to complex - imaginary parts all set to zero.
	//
	// Since 'samples' and 'bins' can be the same buffer, we'll work from the end,
	// so as to not overwrite our data.
	//
	// This unfortunately means we need 2x longer buffer than is the number of samples,
	// and 4x longer than the bin count.
	//
	// That limits our max sample count to 2048 with a 4096 byffer, which yields 1024 bins of FFT.
	// At 4096 Hz, that is 4 Hz per bin.
	//
	// The RFFT function is supposed to solve this better, but it doesn't seem to work properly :(

	for (int i = samp_count - 1; i >= 0; i--) {
		bins[i * 2 + 1] = 0;                   // imaginary
		bins[i * 2] = samples[i] * window[i]; // real
	}

	uint32_t bin_count = samp_count/2;

	arm_cfft_f32(S, bins, 0, true); // bit reversed FFT
	arm_cmplx_mag_f32(bins, bins, bin_count); // get magnitude (extract real values)

	// normalize
	float factor = 1.0f/bin_count;
	for(uint32_t i = 0; i < bin_count; i++) {
		bins[i] *= factor;
	}

//	dbg("FFT done.");
}
