#include "sbmp_control.h"
#include "debug.h"
#include "acquire/acquire_control.h"
#include "utils/timebase.h"
#include "malloc_safe.h"
#include "hw_utils/watchdog.h"
#include "gui/infoled.h"

#include "globals.h"
#include "acquire/processing.h"
#include "gui/colorled.h"

#include "acquire/monitoring.h"

#include "gui/wifistatusled.h"

#include <arm_math.h>

/* This module handles incoming SBMP messages and does the data readout */

// ms - how long the ESP has to read out all the data
#define READOUT_TMEO 200

static void readout_terminate(void);

enum req_format {
	RAW, FFT
};

// --- Data reading ---

static struct rd_struct {
	bool sampling_in_progress; /*!< Sampling or readout in progress (basically a mutex of sorts) */
	bool readout_in_progress;  /*!< Sampling done, but data not yet read out - must remove sesn listener etc. */
	bool readout_initiated; /*!< ESP started reading data - that means our offer went through without error */
	task_pid_t timeout_task_pid; /*!< PID of the readout timeout task */
	uint16_t sesn; /*!< Data readout session number */

	enum req_format format; /*!< Requested data format - post-processing. */

	uint16_t sample_count; /*!< Number of requested samples */

	uint8_t *byte_buffer;  /*!< Buffer ready for readout - can be other than the global buffer in some cases */
	uint32_t len_bytes;    /*!< Buffer length */
	bool buffer_is_mallocd; /*!< If true, buffer will be freed upon termination */

	uint8_t infodata[32]; /*!< Block of data sent with the offer, kept here for retries */
	uint16_t infodata_len;
} rd;

// --- Timeout task ---

static void readout_timeout_task_cb(void *arg)
{
	(void)arg;
	// clean up
	warn("Readout timed out, cleaning up...");
	readout_terminate();
}

static void restart_timeout_task(ms_time_t ms)
{
	abort_scheduled_task(rd.timeout_task_pid);
	rd.timeout_task_pid = schedule_task(readout_timeout_task_cb, NULL, ms, true);
}

static void stop_timeout_task(void)
{
	abort_scheduled_task(rd.timeout_task_pid);
	rd.timeout_task_pid = 0;
}


static void readout_terminate(void)
{
	rd.readout_in_progress = false;

	dbg("Readout terminated.");
	stop_timeout_task();
	sbmp_ep_remove_listener(dlnk_ep, rd.sesn);

	if (rd.buffer_is_mallocd && rd.byte_buffer) free(rd.byte_buffer);
	rd.byte_buffer = NULL;
}

// --- Chunked SBMP readout ---

/**
 * Handle bulk transfer.
 *
 * We will wait for ABORT even after all data is read,
 * so master can re-read some parts if needed.
 *
 * Timeout cloises the session if master fails to do so.
 */
static void bulk_readout_handler(SBMP_Endpoint *ep, SBMP_Datagram *dg, void **obj)
{
	UNUSED_ARG(obj);

	PayloadParser pp;
	switch (dg->type) {
		case DG_BULK_ABORT:
			// master decided to "close the socket"
			info("Data transfer terminated by peer.");
			goto close;

		case DG_BULK_REQUEST:
			rd.readout_initiated = true; // mark that esp started reading - stop the wait/retry loop for offer.

			// master wants some data
			pp = pp_start(dg->payload, dg->length);

			uint32_t offset = pp_u32(&pp);
			uint16_t length = pp_u16(&pp);

			if (length > rd.len_bytes - offset) {
				length = (uint16_t)(rd.len_bytes - offset);
			}

			if (offset > rd.len_bytes) {
				// offset past end of data
				error("Requested bulk offset out of range, aborting.");
				sbmp_bulk_abort(ep, dg->session);
				goto close;
			}

			// feed the dog
			WDG_Feed();
			restart_timeout_task(READOUT_TMEO);
			sbmp_bulk_send_data(ep, rd.byte_buffer + offset, length, dg->session);
			break;
	}

	return;
close:
	readout_terminate();
}


// ---- Data capture ---

static void capt_offer_future_cb(void *unused)
{
	(void)unused;

	bool suc = false;

	dbg("Waiting for esp to start reading...");
	rd.readout_initiated = false;
	// offer the data for readout
	for (int i = 0; i < 3; i++) {
		suc = sbmp_bulk_offer(dlnk_ep, rd.len_bytes, rd.infodata, rd.infodata_len, rd.sesn);
		if (!suc) {
			warn("failed to send offer DG!");
			continue;
		}
		suc = false;

		until_timeout(50) {
			com_poll(data_iface);
			if (rd.readout_initiated) {
				suc = true;
				break;
			}
		}

		if (suc) {
			break;
		}
	}

	if (!suc) {
		warn("Readout failed to start, cleaning up...");
		sbmp_bulk_abort(dlnk_ep, rd.sesn);
		readout_terminate();
	}
}


/** Capture success handler */
static void capture_done_hdlr(uint32_t *buf, uint16_t len, void *xtra)
{
	UNUSED_ARG(buf);
	UNUSED_ARG(len);
	UNUSED_ARG(xtra);

	rd.sampling_in_progress = false; // sampling done: If peer now starts new capture, buffer will be invalidated
	rd.readout_in_progress = true;

	bool suc;

	// FFT if requested

	// we're using the global buffer
	rd.buffer_is_mallocd = false;

	samples_to_milliamps(samp_buf.floats, samp_buf.uints, rd.sample_count);

	// --- At this point, the floats array contains inst. current in milliamps. ---

	PayloadBuilder pb = pb_start(rd.infodata, 32);

	// -- add extra data --
	pb_u32(&pb, rd.sample_count);

	// actual frequency
	float hz = (float)F_CPU / ((TIM1->PSC + 1) * (TIM1->ARR + 1)); // this should be in the sampling module instead
	pb_float(&pb, hz);

	float val;
	uint32_t idx;

	// minimum
	arm_min_f32(samp_buf.floats, rd.sample_count, &val, &idx);
	pb_float(&pb, val);

	// maximum
	arm_max_f32(samp_buf.floats, rd.sample_count, &val, &idx);
	pb_float(&pb, val);

	// RMS
	arm_rms_f32(samp_buf.floats, rd.sample_count, &val);
	pb_float(&pb, val);


	// -- process the payload --
	switch (rd.format) {
		case RAW:
			// send the samples as is (mA)
			rd.byte_buffer = samp_buf.as_bytes; // as bytes
			rd.len_bytes = rd.sample_count * sizeof(uint32_t); // 4 bytes
			break;

		case FFT:
			// Take FFT of the samples. Samples are now in mA
			samples_to_fft(samp_buf.floats, samp_buf.floats, rd.sample_count);

			rd.byte_buffer = samp_buf.as_bytes; // as bytes
			rd.len_bytes = (rd.sample_count / 2) * sizeof(uint32_t); // 4 bytes, HALF the length - FFT halves it!
	}

	restart_timeout_task(READOUT_TMEO); // we are now waiting for client to read the data

	// add a listener for the session
	suc = sbmp_ep_add_listener(dlnk_ep, rd.sesn, bulk_readout_handler, NULL);
	if (!suc) goto fail;

	rd.infodata_len = (uint16_t)pb_length(&pb);

	tq_post(capt_offer_future_cb, NULL);
	return;

fail:
	warn("Readout failed to start, cleaning up...");
	sbmp_bulk_abort(dlnk_ep, rd.sesn);
	readout_terminate();
}

/** Capture fail handler */
static void capture_fail_hdlr(void *xtra)
{
	(void)xtra;

	rd.sampling_in_progress = false;

	// buffer was already freed
	warn("Capture failed, cleaning up...");
	readout_terminate();
}


static void normalize_freq_and_count(uint32_t *frequency, uint16_t *count)
{
	if (*count > SAMP_BUF_LEN) {
		warn("Max count is "STR(SAMP_BUF_LEN)".");
		*count = SAMP_BUF_LEN;
	}

	if (*count == 0) {
		warn("Min count is 1.");
		*count = 1;
	}

	if (*frequency > 5E6) {
		warn("Max freq is 5 MHz.");
		*frequency = 5E6;
	}

	if (*frequency == 0) {
		warn("Min freq is 1 Hz.");
		*frequency = 1;
	}
}

/**
 * @brief Start capture for request of given format
 * @param sesn   : session nr
 * @param format : requested result format
 * @param count  : number of samples
 * @param freq   : sampling frequency
 */
static void start_capture_for_req(uint16_t sesn, enum req_format format, uint16_t count, uint32_t freq)
{
	// populate the info object
	rd.sesn = sesn;
	rd.format = format;
	rd.sample_count = count;

	// coerce to valid range
	normalize_freq_and_count(&freq, &count);

	dbg("%"PRIu32" Hz, %"PRIu32" samples", freq, count);

	// buffer will be malloc'd
	bool suc = start_capture_task(samp_buf.uints, count, freq, capture_done_hdlr, capture_fail_hdlr, NULL);
	if (!suc) {
		// already reading, send abort message back
		error("Capture failed to start.");
		sbmp_bulk_abort(dlnk_ep, sesn);
		return;
	}

	rd.sampling_in_progress = true;
}


/** confirm a request */
static void confirm_request(uint16_t sesn)
{
	sbmp_ep_send_response(dlnk_ep, DG_RQ_CONFIRM, NULL, 0, sesn, NULL);
}

/** Handle request to sample raw */
static void handle_req_raw(SBMP_Datagram *dg)
{
	PayloadParser pp = pp_start(dg->payload, dg->length);
	uint16_t count = pp_u16(&pp);
	uint32_t frequency = pp_u32(&pp);

	if (rd.sampling_in_progress) {
		error("Sampling in progress.");
		goto send_abort;
	}

	if (rd.readout_in_progress) {
		warn("Aborting unfinished readout.");
		readout_terminate();
	}

	confirm_request(dg->session);
	start_capture_for_req(dg->session, RAW, count, frequency);
	return;

send_abort:
	sbmp_bulk_abort(dlnk_ep, dg->session);
}


/** Handle request to measure FFT */
static void handle_req_fft(SBMP_Datagram *dg)
{
	PayloadParser pp = pp_start(dg->payload, dg->length);
	uint16_t count = pp_u16(&pp);
	uint32_t frequency = pp_u32(&pp);

	if (rd.sampling_in_progress) {
		// already reading, send abort message back
		error("Sampling in progress.");
		goto send_abort;
	}

	if (rd.readout_in_progress) {
		warn("Aborting unfinished readout.");
		readout_terminate();
	}

	// validation (generic)
	normalize_freq_and_count(&frequency, &count);

	if (!IS_FFT_LEN(count)) {
		error("Invalid sample count %"PRIu16" for FFT.", count);
		goto send_abort;
	}

	confirm_request(dg->session);
	start_capture_for_req(dg->session, FFT, count, frequency);
	return;

send_abort:
	sbmp_bulk_abort(dlnk_ep, dg->session);
}


// --- store healthy reference ---

/** Waveform captured for reference */
static void healthy_cap_done_hdlr(uint32_t* buf, uint16_t count, void *arg)
{
	(void)buf;
	(void)count;
	(void)arg;

	rd.sampling_in_progress = false;

	// We now have reference in the buffer

	bool suc = rpt_store_reference();

	sbmp_ep_send_response(dlnk_ep, suc ? DG_SUCCESS : DG_FAILURE, NULL, 0, rd.sesn, NULL);
}

/** Error capturing waveform for rereference */
static void healthy_cap_fail_hdlr(void *arg)
{
	(void)arg;

	rd.sampling_in_progress = false;
	error("FAILED TO STORE \"HEALTHY\" REFERENCE!");

	sbmp_ep_send_response(dlnk_ep, DG_FAILURE, NULL, 0, rd.sesn, NULL);
}

/** Request to save current waveform as reference */
static void handle_store_ref_req(SBMP_Datagram *dg)
{
	bool suc = start_capture_task(samp_buf.uints,
								  HEALTH_REF_SAMP_LEN,
								  HEALTH_REF_SAMPLE_RATE,
								  healthy_cap_done_hdlr,
								  healthy_cap_fail_hdlr,
								  NULL);
	if (!suc) {
		error("Capture failed to start.");
		return;
	}

	rd.sesn = dg->session;
	rd.sampling_in_progress = true;
}

// -------------------------------

/** Waveform captured for reference */
static void refcomp_done_hdlr(uint32_t* buf, uint16_t count, void *arg)
{
	(void)buf;
	(void)count;
	(void)arg;

	rd.sampling_in_progress = false;

	// We now have reference in samp_buf

	reference_info info;
	rpt_process_captured_vector(&info); // convert vector to milliamps (float), and also get rms.

	float dist = rpt_calc_vector_dist();
	dbg("dist = %.2f", dist);

	sbmp_ep_start_response(dlnk_ep, DG_SUCCESS, 8, rd.sesn);

	sbmp_ep_send_float(dlnk_ep, dist);
	sbmp_ep_send_float(dlnk_ep, info.rms);
}

/** Error capturing waveform for rereference */
static void refcomp_fail_hdlr(void *arg)
{
	(void)arg;

	rd.sampling_in_progress = false;
	error("FAILED TO STORE REFERENCE!");
	sbmp_ep_send_response(dlnk_ep, DG_FAILURE, NULL, 0, rd.sesn, NULL);
}

/** Request to save current waveform as reference */
static void handle_compare_ref_req(SBMP_Datagram *dg)
{
	if (!rpt_is_ref_stored()) {
		error("Reference is not stored, cannot measure deviation.");
		sbmp_ep_send_response(dlnk_ep, DG_FAILURE, NULL, 0, rd.sesn, NULL);
		return;
	}


	bool suc = start_capture_task(samp_buf.uints,
								  HEALTH_REF_SAMP_LEN,
								  HEALTH_REF_SAMPLE_RATE,
								  refcomp_done_hdlr,
								  refcomp_fail_hdlr,
								  NULL);
	if (!suc) {
		error("Capture failed to start.");
		return;
	}

	rd.sesn = dg->session;
	rd.sampling_in_progress = true;
}


// -------------------------------

static task_pid_t esp_not_connected_pid;

static void esp_not_connected_timer_cb(void *unused)
{
	(void)unused;
	memset(&WifiStatus, 0, sizeof(struct wifistatus));
	wifistatus_onchange();
}

void wifistatus_monitoring_init(void)
{
	esp_not_connected_pid = add_periodic_task(esp_not_connected_timer_cb, NULL, 3000, true);
}


/** ESP sent it's version to us */
static void handle_wifi_status_dg(SBMP_Datagram *dg)
{
	reset_periodic_task(esp_not_connected_pid); // clear the "disconnected watchdog"

	PayloadParser pp = pp_start(dg->payload, dg->length);

	//-----------------------------------------
	// opmode | 1-Client, 2-SoftAP, 3-STA+AP
	// Station connected
	// 1 = WPS is in progress
	// 1 = Reporting is enabled
	//-----------------------------------------

	uint8_t opmode = pp_u8(&pp);
	bool sta_connected = pp_u8(&pp);
	bool wps_active = pp_u8(&pp);
	bool reporting_enabled = pp_u8(&pp);

	if (WifiStatus.opmode != opmode
			|| WifiStatus.sta_connected != sta_connected
			|| WifiStatus.wps_active != wps_active
			|| WifiStatus.reporting_enabled != reporting_enabled) {

		dbg("WiFi status change: %s, Sta-conn %d, WPS %d, Reporting %d",
			opmode == 1 ? "Client" : opmode == 2 ? "SoftAP" : "STA+AP",
			sta_connected,
			wps_active,
			reporting_enabled);

		WifiStatus.opmode = opmode;
		WifiStatus.sta_connected = sta_connected;
		WifiStatus.wps_active = wps_active;
		WifiStatus.reporting_enabled = reporting_enabled;

		wifistatus_onchange();
	}
}


/** ESP asked for our version, send it back as 3 bytes */
static void handle_stmversion_request_dg(SBMP_Datagram *dg)
{
	if (sbmp_ep_start_response(dlnk_ep, DG_SUCCESS, 3, dg->session)) {
		sbmp_ep_send_u8(dlnk_ep, FW_VERS_MAJOR);
		sbmp_ep_send_u8(dlnk_ep, FW_VERS_MINOR);
		sbmp_ep_send_u8(dlnk_ep, FW_VERS_PATCH);
	}
}

/** Handle incoming SBMP datagram - registered in datalink. */
void dlnk_rx(SBMP_Datagram *dg)
{
	switch (dg->type) {
		case DG_REQUEST_RAW:
			handle_req_raw(dg);
			break;

		case DG_REQUEST_FFT:
			handle_req_fft(dg);
			break;

		case DG_REQUEST_STORE_REF:
			handle_store_ref_req(dg);
			break;

		case DG_REQUEST_COMPARE_REF:
			handle_compare_ref_req(dg);
			break;

		case DG_WIFI_STATUS:
			handle_wifi_status_dg(dg);
			break;

		case DG_REQUEST_STM_VERSION:
			handle_stmversion_request_dg(dg);
			break;
	}
}

// --------------------------------

void wificontrol_start_wps(void *unused)
{
	(void)unused;

	info("Request WPS start");
	sbmp_ep_send_message(dlnk_ep, DG_WPS_START, NULL, 0, NULL, NULL);

	WifiStatus.wps_preparing = true;
	wifistatus_onchange();
}


void wificontrol_set_ap(void *unused)
{
	(void)unused;

	info("Request AP mode");
	sbmp_ep_send_message(dlnk_ep, DG_SETMODE_AP, NULL, 0, NULL, NULL);
}


