#include "sampling.h"
#include "hw_utils/all.h"
#include "com/debug.h"

#include "utils/timebase.h"
#include "utils/mutex.h"


// --- Priv protos ----------

/** Lock the mutex */
static bool acq_take_lock(void);
/** Release the mutex */
static void acq_release_lock(void);

/** Set up TIM1 (ADC trigger for batch capture) */
static void conf_adc(void);
/** Set up ADC1 */
static void conf_tim(void);

/** Prepare DMA for bratch capture */
static void dma_prepare(void *buffer, uint16_t count, enum DMADataWidth buf_width);
/** Kick off the batch capture */
static bool dma_start(void);
/** Clean up after DMA */
static void dma_cleanup(void);


// --- Vars n types ----------

/* Common settings for the ACQ module */
static const uint8_t ACQ_SAMPLE_TIME = ADC_SampleTime_61Cycles5;
static const uint32_t ADC1_TIM1_TRG_EVT = ADC_ExternalTrigConvEvent_9;

/** ACQ mutex - locked when ADC is in use */
static mutex_t acq_mutex;

static struct current_dma_struct {
	void *buffer;                   ///< DMA target buffer
	enum DMADataWidth buffer_width; ///< data width of the buffer
	uint16_t length;                ///< buffer data count
	bool running;                   ///< indicates that DMA is in progress
} current_dma;


// --- Funcs -----------------


static void resolve_presc_for_freq(uint32_t f_desired, uint16_t *prescaller, uint16_t *period)
{
	uint32_t div_desired = F_CPU / f_desired;

	uint16_t presc = 1;
	while (div_desired > 0xFFFF) {
		div_desired >>= 1;
		presc <<= 1;
	}
	*period = (uint16_t)(div_desired - 1);
	*prescaller = presc - 1;
}


/** Initialize the hardware resources for data acquisition */
void acq_init(void)
{
	// ---- RCC --------------

	// Enable clock for the ADC
	RCC_AHBPeriphClockCmd(RCC_AHBENR_ADC12EN, ENABLE); // ADC1
	RCC_AHBPeriphClockCmd(RCC_AHBENR_DMA1EN, ENABLE); // DMA1
	RCC_APB2PeriphClockCmd(RCC_APB2ENR_TIM1EN, ENABLE); // TIM1

	// set clock for ADC1_2
	RCC_ADCCLKConfig(RCC_ADC12PLLCLK_Div1);

	// ADC
	conf_adc();

	// Triggering timer
	conf_tim();
}


/** Configure the ADC */
static void conf_adc(void)
{
	ADC_DeInit(ADC1); // Reset ADC1_2

	// Common config for ADC1_2

	ADC_CommonInitTypeDef adcc_cfg;

	adcc_cfg.ADC_Mode = ADC_Mode_Independent;
	adcc_cfg.ADC_Clock = ADC_Clock_AsynClkMode;
	adcc_cfg.ADC_DMAAccessMode = ADC_DMAAccessMode_Disabled; // only for multi-DMA
	adcc_cfg.ADC_DMAMode = ADC_DMAMode_OneShot;
	adcc_cfg.ADC_TwoSamplingDelay = 0;

	ADC_CommonInit(ADC1, &adcc_cfg);


	// ADC1 config
	ADC_InitTypeDef adc_cfg;

	adc_cfg.ADC_ContinuousConvMode = DISABLE;
	adc_cfg.ADC_Resolution = ADC_Resolution_12b;
	adc_cfg.ADC_ExternalTrigConvEvent = ADC_ExternalTrigConvEvent_9; // TIM1_TRGO
	adc_cfg.ADC_ExternalTrigEventEdge = ADC_ExternalTrigEventEdge_None;// <- no edge, disabled
	adc_cfg.ADC_DataAlign = ADC_DataAlign_Right;
	adc_cfg.ADC_OverrunMode = DISABLE;
	adc_cfg.ADC_AutoInjMode = DISABLE;
	adc_cfg.ADC_NbrOfRegChannel = 1;

	ADC_Init(ADC1, &adc_cfg);

	// Select channel
	const int chan = 1;
	const int rank = 1;
	ADC_RegularChannelSequencerLengthConfig(ADC1, 1); // not needed
	ADC_RegularChannelConfig(ADC1, chan, rank, ACQ_SAMPLE_TIME);

	// Enable the ADC voltage regulator
	ADC_VoltageRegulatorCmd(ADC1, ENABLE);
	delay_us(20); // wait for regulator to stabilize (overkill delay)

	// perform ADC calibration
	adc_calibrate(ADC1);
}


/** Configure the ADC triggering timer */
static void conf_tim(void)
{
	TIM_DeInit(TIM1); // Reset TIM1

	TIM_TimeBaseInitTypeDef tim_cfg;

	tim_cfg.TIM_Period = (F_CPU / 4096UL); // attention! Must not exceed 65535
	tim_cfg.TIM_Prescaler = 0;
	tim_cfg.TIM_ClockDivision = TIM_CKD_DIV1;
	tim_cfg.TIM_CounterMode = TIM_CounterMode_Up;
	tim_cfg.TIM_RepetitionCounter = 0x0000;

	TIM_TimeBaseInit(TIM1, &tim_cfg);

	// Configure event generation
	TIM_SelectOutputTrigger(TIM1, TIM_TRGOSource_Update);
}


/** Lock the ACQ mutex */
static bool acq_take_lock(void)
{
	bool suc = true;

	if (!mutex_lock(&acq_mutex, 200)) {
		suc = false;
		error("Failed to get access to ADC.");
	}

	return suc;
}


/** Unlock the ACQ mutex */
static void acq_release_lock(void)
{
	mutex_unlock(&acq_mutex);
}


/** Configure DMA1_Ch1 for reading from ADC */
static void dma_prepare(void *buffer, uint16_t count, enum DMADataWidth buf_width)
{
	DMA_DeInit(DMA1_Channel1); // reset the channel

	// Enable ADC<-TRGO evt 9
	ADC_ExternalTriggerConfig(ADC1, ADC1_TIM1_TRG_EVT, ADC_ExternalTrigEventEdge_RisingEdge); // <-- rising

	// convert number of bits to a bitmask for the register
	uint32_t mem_data_size = 0;
	switch (buf_width) {
		case DMADataWidth_8: mem_data_size = DMA_MemoryDataSize_Byte; break;
		case DMADataWidth_16: mem_data_size = DMA_MemoryDataSize_HalfWord; break;
		case DMADataWidth_32: mem_data_size = DMA_MemoryDataSize_Word; break;
	}

	DMA_InitTypeDef dma_cfg;

	dma_cfg.DMA_PeripheralBaseAddr = (uint32_t) & (ADC1->DR);
	dma_cfg.DMA_MemoryBaseAddr = (uint32_t) buffer;
	dma_cfg.DMA_DIR = DMA_DIR_PeripheralSRC; // from peripheral
	dma_cfg.DMA_BufferSize = count;
	dma_cfg.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dma_cfg.DMA_MemoryInc = DMA_MemoryInc_Enable; // increment memory
	dma_cfg.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord; // 16-bit DR on DAC
	dma_cfg.DMA_MemoryDataSize = mem_data_size;
	dma_cfg.DMA_Mode = DMA_Mode_Normal; // normal / circular
	dma_cfg.DMA_Priority = DMA_Priority_High; // high priority -> could get better accuracy
	dma_cfg.DMA_M2M = DMA_M2M_Disable;

	DMA_Init(DMA1_Channel1, &dma_cfg);

	// enable interrupt on complete & error
	DMA_ITConfig(DMA1_Channel1, (DMA1_IT_TC1 | DMA_IT_TE), ENABLE);
}


/** Start DMA1_Ch1 from ADC. The DMA must already be set up. */
static bool dma_start(void)
{
	// Perform ADC calibration (also turns the ADC on)
	if (!adc_calibrate(ADC1)) {
		error("ADC failed to calibrate.");
		return false;
	}

	// adc is turned on by the calib routine

	ADC_StartConversion(ADC1);

	// Enable DMA requests from ADC
	ADC_DMACmd(ADC1, ENABLE);

	// Enable DMA
	DMA_Cmd(DMA1_Channel1, ENABLE);

	// Enable DMA TC irq
	NVIC_EnableIRQ(DMA1_Channel1_IRQn);

	// Start counter
	TIM_Cmd(TIM1, ENABLE);

	// Start the first convertion
	ADC_StartConversion(ADC1);

	return true;
}


/** Clean-up after the ADC DMA ended. */
static void dma_cleanup(void)
{
	// Disable the IRQ
	NVIC_DisableIRQ(DMA1_Channel1_IRQn);

	// turn off the channel
	DMA_Cmd(DMA1_Channel1, DISABLE);

	// Disable ADC dma requests
	ADC_DMACmd(ADC1, DISABLE);

	// turn off the counter
	TIM_Cmd(TIM1, DISABLE);

	// Disable ext trig
	ADC_ExternalTriggerConfig(ADC1, ADC1_TIM1_TRG_EVT, ADC_ExternalTrigEventEdge_None); // <-- No edge = disable

	// Make sure ADSTART is cleared.
	ADC_StopConversion(ADC1);

	// Turn off the ADC
	adc_off(ADC1);
}


/** Check if DMA in progress */
bool acq_dma_in_progress(void)
{
	return current_dma.running;
}


/** Read single sample from the ADC, if DMA not in progress. */
bool acq_one_sample(uint16_t *dest)
{
	bool suc = false;

	if (!acq_take_lock()) {
		return false; // abort
	}

	if (!adc_calibrate(ADC1)) {
		error("ADC calibration failed.");
		goto done;
	}

	ADC_StartConversion(ADC1);

	// wait for completion
	until_timeout(20) {
		if (ADC1->ISR & ADC_ISR_EOS) {
			*dest = (uint16_t)ADC1->DR;
			suc = true;
			goto done;
		}
	}

	error("ADC read timeout.");

done:
	acq_release_lock();
	return suc;
}


/** Start the DMA batch acquisition. Locks the mutex. */
bool acq_start_dma_capture(void *target, uint16_t count, uint32_t frequency, enum DMADataWidth width)
{
	if (!acq_take_lock()) {
		return false; // abort
	}

	dma_prepare(target, count, width);

	// Configure the timer
	uint16_t psc, arr;
	resolve_presc_for_freq(frequency, &psc, &arr);
	TIM1->PSC = psc;
	TIM1->ARR = arr;

	// store info for IRQ
	current_dma.buffer = target;
	current_dma.length = count;
	current_dma.buffer_width = width;
	current_dma.running = true;

	dbg("DMA prepared, starting...");

	if (dma_start()) {
		dbg("DMA started!");

		return true; // started, mutex will be freed in the IRQ
	}

	// --- Error ---

	error("DMA failed to start.");

	current_dma.running = false;

	// clean up, disable resources
	dma_cleanup();
	// free the mutex
	acq_release_lock();

	return false;
}


/**
 * @brief DMA1 Channel1 IRQ handler - ACQ batch capture complete.
 */
void DMA1_Channel1_IRQHandler(void)
{
	bool is_DMA1Ch1_tc = DMA_GetFlagStatus(DMA1_FLAG_TC1);
	bool is_DMA1Ch1_te = DMA_GetFlagStatus(DMA1_FLAG_TE1);

	if (is_DMA1Ch1_tc || is_DMA1Ch1_te) {

		// Stop DMA, disable periphs
		dma_cleanup();

		if (is_DMA1Ch1_tc) {
			DMA_ClearFlag(DMA1_FLAG_TC1);

			dbg_from_ISR(LEVEL_INFO, "DMA completed.");

			// send success event
			Event evt;
			evt.type = EVT_ACQ_DMA_COMPLETE;
			evt.num_a = current_dma.length;
			evt.num_b = current_dma.buffer_width;
			evt.data_ptr = current_dma.buffer;
			eq_post(&evt);

		} else {
			// error event
			DMA_ClearFlag(DMA1_FLAG_TE1);

			dbg_from_ISR(LEVEL_ERROR, "DMA transfer failed.");

			// send error event
			Event evt;
			evt.type = EVT_ACQ_DMA_FAILED;
			evt.data_ptr = current_dma.buffer;
			eq_post(&evt);
		}

		// unmark that it's running
		current_dma.running = false;

		// Unlock the mutex
		acq_release_lock();
	}
}
